' Lets be amazed!
' "3D" Maze Runner
' written by vegipete, June 2020
' Based on "monstermaze" found on the internet.

WIDTH = 15
HEIGHT = 11
dim m$(HEIGHT) length WIDTH
m$( 1) = "###############"
m$( 2) = "#   # #     # #"
m$( 3) = "### # # # ### #"
m$( 4) = "#   # # #   # #"
m$( 5) = "# # # # ### # #"
m$( 6) = "# # #   # # # #"
m$( 7) = "# # # ### # # #"
m$( 8) = "# #   #       #"
m$( 9) = "# ### # ##### #"
m$(10) = "#     #     #  "
m$(11) = "###############"

' player starting position, orientation
playerX = 2
playerY = 2
playerD = 2 ' 1=up, 2=right, 3=down, 1=left
DIM INTEGER deltaX(4) = (0, 0, 1, 0,-1)
DIM INTEGER deltaY(4) = (0,-1, 0, 1, 0)

' colour definitions
GROUND = &H60FF60
SKY = &H6060FF
WALL = &HA0A0A0     ' wall seen flat on
WALL2 = &H404040    ' wall seen at angle

mode 2,8        ' 640 x 400, 8 bit colour depth

' load the side wall images
page write 2
load png "side0123456.png",0,0,15

page write 1

' draw starting view
DrawMaze

do
  update = 0
  k = asc(INKEY$) ' grab any key press
  if k = 128 then ' UP arrow = forwards
    nx = playerX + deltaX(playerD)
    ny = playerY + deltaY(playerD)
    if mid$(m$(ny),nx,1) = " " then
      playerX = nx
      playerY = ny
      update = 1
    endif
  endif
  if k = 129 then ' DOWN arrow = backwards
    nx = playerX - deltaX(playerD)
    ny = playerY - deltaY(playerD)
    if mid$(m$(ny),nx,1) = " " then
      playerX = nx
      playerY = ny
      update = 1
    endif
  endif
  if k = 130 then ' LEFT arrow = turn left 90
    playerD = playerD - 1
    if playerD = 0 then playerD = 4
    update = 1
  endif
  if k = 131 then ' RIGHT arrow = turn right 90
    playerD = playerD + 1
    if playerD = 5 then playerD = 1
    update = 1
  endif
  if k = 115 then ' s save graphic image as bmp
    save image "screenshot",0,0,400,400
  endif
  
  ' only redraw maze if something has changed
  if update then
    DrawMaze
    update = 0
  endif
loop

end

'################################################################
' Draw the maze as seen by the player
' standing at (playerX,playerY) and facing direction playerD
' (where 1=up, 2=right, 3=down, 1=left)
' While the image looks 3 dimensional, it isn't really. Instead,
' it is simulated by just drawing carefully generated images
' such that nearer wall images cover over smaller more distant
' wall images. Fortunately, the CMM2 is fast enough that drawing
' a bunch of wall images that just get covered over again by
' nearer walls isn't a significant performance hit.
Sub DrawMaze

local vx,vy,sx,sy,side

t = timer

page write 1
cls

' draw background, distant wall
box 0,  0,400,200,1,SKY,SKY
box 0,200,400,200,1,GROUND,GROUND

' what can we see?
side = 1
if playerD = 2 or playerD = 4 then side = -1

' find walls from farther to nearer
for dist = 6 to 0 step -1
  ' view a cell some distance in front
  vx = playerX + deltaX(playerD) * dist
  vy = playerY + deltaY(playerD) * dist
  if vx > 0 and vx <= WIDTH and vy > 0 and vy <= HEIGHT then
    ' force a wall beyond vision distance
    if dist = 6 then box 190,199,20,2,1,WALL 
    ' check for wall to the left
    sx = vx + deltaY(playerD) * side
    sy = vy + deltaX(playerD) * side
    if mid$(m$(sy),sx,1) <> " " then DrawLeftWall(dist)
    ' check for wall to the right
    sx = vx - deltaY(playerD) * side
    sy = vy - deltaX(playerD) * side
    if mid$(m$(sy),sx,1) <> " " then DrawRightWall(dist)
    ' check for wall ahead
    if mid$(m$(vy),vx,1) <> " " then DrawFarWall(dist)
  endif
next dist

page copy 1 to 0

page write 0
print @(410,0) "Time to draw screen (ms)
print @(410,20) timer - t

end sub

'################################################################
' Subroutines to draw walls at various distances
' either to left, to the right or straight ahead
'################################################################
sub DrawLeftWall(dist)

if dist > 6 then exit sub

if dist = 6 then blit 175,386,175,197,25,6,2,4 ' respect transparency
if dist = 5 then blit 175,374,175,194,22,12,2,4
if dist = 4 then blit 150,350,150,188,44,24,2,4
if dist = 3 then blit 100,300,100,175,88,50,2,4
if dist = 2 then blit 100,200,100,150,75,100,2,4
if dist = 1 then blit 0,0,0,100,150,200,2,4
if dist = 0 then blit 200,0,0,0,100,400,2,4

end sub

'################################################################
' Right side walls are mirror images of matching left side walls.
' Conveniently, blit can mirror!
sub DrawRightWall(dist)

if dist > 6 then exit sub

if dist = 6 then blit 175,386,200,197,25,6,2,5  ' mirrored
if dist = 5 then blit 175,374,203,194,22,12,2,5  ' mirrored
if dist = 4 then blit 150,350,206,188,44,24,2,5  ' mirrored
if dist = 3 then blit 100,300,212,175,88,50,2,5  ' mirrored
if dist = 2 then blit 100,200,225,150,75,100,2,5 ' mirrored
if dist = 1 then blit 0,0,250,100,150,200,2,5 ' mirrored
if dist = 0 then blit 200,0,300,0,100,400,2,5 ' mirrored

end sub

'################################################################
sub DrawFarWall(dist)

if dist > 6 then exit sub

if dist = 6 then box 197,197,  6,  6,1,WALL,WALL
if dist = 5 then box 194,194, 12, 12,1,WALL,WALL
if dist = 4 then box 188,188, 24, 24,1,WALL,WALL
if dist = 3 then box 175,175, 50, 50,1,WALL,WALL
if dist = 2 then box 150,150,100,100,1,WALL,WALL
if dist = 1 then box 100,100,200,200,1,WALL,WALL

end sub
